# Copyright (C) 2018 Matthew L. Schwartz - all rights reserved
# Not for redistribution

import SapTrap_Functions as STF
import operator


class DNA:

    def __init__(self, sequence):
        self.sequence = sequence
        self.dictBasePairs = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A',
                         'a': 't', 'c': 'g', 'g': 'c', 't': 'a',
                         'N': 'N', 'n': 'n', '-': '-', ' ': ' '}

        return None

    def __len__(self):
        return len(self.sequence)

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < len(self.sequence):
            result = self.sequence[self.n]
            self.n += 1
            return result
        else:
            raise StopIteration

    def __str__(self):
        return self.sequence

    def __add__(self, other):
        return self.sequence + other

    def __getitem__(self, slice_notation):
        return self.sequence[slice_notation]

    def reverse(self):
        """ returns sequence in reverse order """
        return DNA(self.sequence[::-1])

    def complement(self):
        """generates complement of sequence"""
        complement = ''
        for base in self.sequence:
            complement += self.dictBasePairs[base]
        return DNA(complement)

    def reverse_complement(self):
        """generates reverse-complement of sequence"""
        reverse_sequence = self.reverse()
        complement = ''
        for base in reverse_sequence:
            complement += self.dictBasePairs[base]
        return DNA(complement)

    def find(self, sequence, start):
        return self.sequence.find(sequence, start)

    def lower(self):
        return DNA(self.sequence.lower())

    def upper(self):
        return DNA(self.sequence.upper())

    def islower(self):
        if self.sequence.islower():
            return True
        else:
            return False

    def isupper(self):
        if self.sequence.isupper():
            return True
        else:
            return False


class gRNABindingSite(DNA):

    def __init__(self, start, sequence, direction, guide_specification):
        DNA.__init__(self, sequence)
        self.start = start
        self.direction = direction
        self.guide_specification = guide_specification
        if self.direction == 'sense':
            self.cutsite = self.start + self.guide_specification.cut_site
        if self.direction == 'antisense':
            self.cutsite = self.start - self.guide_specification.cut_site
        self.distance = None
        self.flags = self.warnings()
        PAM_length = len(self.guide_specification.PAM)
        if self.guide_specification.PAM_location == 'RIGHT':
            self.binding_site = self.sequence[:-PAM_length]
        elif self.guide_specification.PAM_location == 'LEFT':
            self.binding_site = self.sequence[PAM_length:]

        return None

    def aligned_binding_site(self):
        """ for printing gRNA aligned to EditSite, adds needed spacers to string """
        if self.direction == 'sense':
            pad_length = self.start
            print_sequence = self.sequence
        elif self.direction == 'antisense':
            pad_length = self.start - len(self.sequence)
            print_sequence = self.reverse_complement()
        print_padding = ''
        for i in range(pad_length):
            print_padding += ' '
        return print_padding + str(print_sequence)

    def aligned_pam(self):
        """ for printing PAM string aligned to EditSite, adds spacers to string """
        pad_length = self.start
        if self.guide_specification.PAM_location == 'RIGHT':
            print_PAM = DNA(self.sequence[0 - self.guide_specification.bind_start:])
            if self.direction == 'sense':
                pad_length -= self.guide_specification.bind_start
            elif self.direction == 'antisense':
                pad_length -= len(self.sequence)
                print_PAM = print_PAM.reverse_complement()
        if self.guide_specification.PAM_location == 'LEFT':
            print_PAM = DNA(self.sequence[:len(self.guide_specification.PAM)])
            if self.direction == 'antisense':
                pad_length -= self.guide_specification.bind_start
                print_PAM = print_PAM.reverse_complement()
        print_padding = ''
        for i in range(pad_length):
            print_padding += ' '
        pam_string = print_padding + str(print_PAM)
        for i in range(len(self.sequence)):
            pam_string += ' '
        pam_string = pam_string[:self.cutsite - 1] + '/\\' + pam_string[self.cutsite + 1:]
        gap = len(str(self.distance)) + 2
        distance = '(' + str(self.distance) + ')'
        if self.guide_specification.PAM_location == 'RIGHT':
            if self.direction == 'sense':
                pam_string = pam_string[:self.start] + distance + pam_string[self.start + gap:]
            if self.direction == 'antisense':
                pam_string = pam_string[:self.start - gap] + distance
        if self.guide_specification.PAM_location == 'LEFT':
            if self.direction == 'sense':
                pam_string = pam_string[:self.cutsite - gap - 2] + distance + pam_string[self.cutsite - 2:]
            if self.direction == 'antisense':
                pam_string = pam_string[:self.cutsite + 2] + distance + pam_string[self.cutsite + gap + 2:]
        return pam_string


    def aligned_guide_flags(self):
        aligned_flags = []
        if self.flags != []:
            flag_pad = 0
            flag_print_padding = ''
            guide_string = self.aligned_binding_site()
            while guide_string[flag_pad] == ' ':
                flag_pad += 1
            for i in range(flag_pad):
                flag_print_padding += ' '
            for flag in self.flags:
                aligned_flags.append(flag_print_padding + flag)
        return aligned_flags


    def warnings(self):
        flags = []
        sequence = str(self.sequence)
        if 'gctcttc' in sequence.lower():
            flags.append('gRNA sequence contains SapI site. Sequence incompatible with SapTrap assembly.')
        elif 'gaagagc' in sequence.lower():
            flags.append('gRNA sequence contains SapI site. Sequence incompatible with SapTrap assembly.')
        if 'tttttt' in sequence.lower():
            flags.append('gRNA sequence contains polIII terminator. Sequence incompatible with production in vivo.')
        if self.guide_specification.nuclease == 'Cas9':
            if sequence[19].lower() == 'c':
                flags.append('gRNA binding site ends in C base. Such guides exhibit low activity in C. elegans.')
        return flags


class GuideSpecification:
    """ PAM starts at position 0.
        PAM is bases of PAM, for example NGG for Cas9
        bind_start, bind_end, cut_site are signed integers specifying distance from PAM start to feature
        PAM_location is 'RIGHT' or 'LEFT', where PAM is relative to binding site """
    def __init__(self, PAM, bind_start, bind_end, cut_site, PAM_location, nuclease):
        self.PAM = PAM
        self.bind_start = bind_start
        self.bind_end = bind_end
        self.PAM_location = PAM_location
        if self.PAM_location == 'RIGHT':
            self.cut_site = 0 - self.bind_start + cut_site
        else:
            self.cut_site = cut_site
        self.nuclease = nuclease

        return None

    def minimum_location(self):
        """ returns left-most linear coordinate of guide binding site """
        if self.PAM_location == 'RIGHT':
            return min(self.bind_start, self.bind_end)
        elif self.PAM_location == 'LEFT':
            return 0

    def maximum_location(self):
        """ returns right-most linear coordinate of guide binding site """
        if self.PAM_location == 'RIGHT':
            return len(self.PAM) - 1
        elif self.PAM_location == 'LEFT':
            return max(self.bind_start, self.bind_end)


class EditSite:

    def __init__(self, sequence, guide_specification, insert_site=None):
        """ Constructor is moved to self.reset method. This allows me to keep only one instance of Edit_site.
                    Whenever user wants to switch sites, the extant instance is effectively re-initialized by self.reset"""
        self.reset(sequence, guide_specification, insert_site)


    def reset(self, sequence, guide_specification, insert_site=None):
        """ Edit_site constructor. Requires only DNA sequence populate attributes.
            This program uses extensive getter and setter methods on Edit_site attributes.
            Program builds final output by sequentially adding and modifying data stored in attributes.
            """
        self.sequence = DNA(sequence)
        self.mutated_sequence = DNA(sequence)
        self.temporary_sequence_hold = DNA(sequence)
        self.guide_specification = guide_specification
        if insert_site is None:
            self.insert_site = int(len(sequence)/2)
        else:
            self.insert_site = insert_site
        self.original_insert_site = self.insert_site
        self.show_insert_site = False
        self.mutation_site = None
        self.frame = self.guess_frame()
        self.gRNA_list = self.sort_guide_list(self.find_gRNA_sites())
        self.gRNA = None
        self.sapI_flag = 0
        self.base_translation = self.translate(self.sequence)
        self.translation_difference_string = self.difference_string('Translation')
        self.mutant_translation = self.translate(self.mutated_sequence)
        self.pointer_string = None
        self.DNA_difference_string = self.difference_string('DNA')
        self.gRNA_string = None
        self.PAM_string = None
        self.guide_flags = []
        self.SapI_string = None
        self.jobname = ''
        self.sap_option = 'SapTrap'
        self.hUS_target_length = 57
        self.hDS_target_length = 57
        self.homology_arm_minimum_length = 32
        self.tag_type = None
        self.hUS = None
        self.hDS = None
        self.repair_template = None
        self.homology_arm_flags = []
        self.large_deletion_range = None
        # should add back in translations, difference strings, etc and calls to make them,
        # reference in display getter
        # all the on site calculation is bogging the program down

    def reverse(self, sequence):
        return str(DNA(sequence).reverse())

    def reverse_complement(self, sequence):
        return str(DNA(sequence).reverse_complement())

    def incorporate_large_deletion(self, deletion_region):
        """ Called during import if <..> used to indicate user wants to delete a region of the genome
            replaces bases enclosed by <..> with -'s in mutated_sequence """
        deletion_string = self.mutated_sequence[:deletion_region[0]]
        for i in range(deletion_region[0], deletion_region[1]):
            deletion_string += '-'
        self.large_deletion_range = [deletion_region[0], deletion_region[1]]
        deletion_string += self.mutated_sequence[deletion_region[1]:]
        self.mutated_sequence = deletion_string

        return 0

    def find_gRNA_sites(self):
        """ Find guide RNA binding sites within a DNA sequence
            Input DNA sequence object and guide_specification object
            Returns list of all guide binding sites (guide_binding_site objects)"""
        guide_list = []
        search_sequence = self.sequence.lower()
        PAMs = (STF.degenerate_sequence_generator(self.guide_specification.PAM, 'lower'))
        if self.guide_specification.PAM_location == 'RIGHT':
            search_start = 0 - self.guide_specification.minimum_location()
        else:
            search_start = 0
        if self.guide_specification.PAM_location == 'LEFT':
            search_end = len(self.sequence) - self.guide_specification.maximum_location()
        else:
            search_end = len(self.sequence)
        for PAM in PAMs:
            local_search_start = search_start
            while PAM in search_sequence[local_search_start: search_end]:
                guide_start = search_sequence.find(PAM, local_search_start) + self.guide_specification.minimum_location()
                guide_end = search_sequence.find(PAM, local_search_start) + self.guide_specification.maximum_location()
                guide_list.append(gRNABindingSite(guide_start, self.sequence[guide_start: guide_end + 1],
                                                      'sense', self.guide_specification))
                local_search_start = search_sequence.find(PAM, local_search_start) + 1
        rc_sequence = self.sequence.reverse_complement()
        search_sequence = rc_sequence.lower()
        for PAM in PAMs:
            local_search_start = search_start
            while PAM in search_sequence[local_search_start: search_end]:
                guide_start = search_sequence.find(PAM, local_search_start) + self.guide_specification.minimum_location()
                guide_end = search_sequence.find(PAM, local_search_start) + self.guide_specification.maximum_location()
                guide_list.append(
                    gRNABindingSite(len(self.sequence) - guide_start, rc_sequence[guide_start: guide_end + 1],
                                        'antisense', self.guide_specification))
                local_search_start = search_sequence.find(PAM, local_search_start) + 1

        return guide_list

    def sort_guide_list(self, guide_list):
        """This function will take a list of guides and assign keys in order of distance between cut and insert"""
        sorted_guide_list = {}
        count = 1
        for gRNA in (sorted(guide_list, key=operator.attrgetter('cutsite'))):
            sorted_guide_list[count] = gRNA
            gRNA.index = count
            count += 1

        return sorted_guide_list

    def set_gRNA(self, gRNA_index):
        """ input is a key in the gRNA library
            No output ; sets gRNA choice, gRNA and PAM display strings"""
        self.gRNA = self.gRNA_list[gRNA_index]
        self.gRNA_string = self.gRNA.aligned_binding_site()
        self.PAM_string = self.gRNA.aligned_pam()
        self.guide_flags = self.gRNA.aligned_guide_flags()
        return 0

    def translate(self, sequence):
        return STF.translate_upper_case(sequence, self.frame, condensed=False)

    def difference_string(self, type):
        """ Input is one of two strings: 'DNA' or 'Translation'
            Operates on Edit_site attributes to produce string highlighting differences between WT and mutant
            Output a string with spaces for identical bases, '*' for differences """
        incorporation_warning = False
        if type == 'DNA':
            string_1 = self.sequence
            string_2 = self.mutated_sequence
        elif type == 'Translation':
            string_1 = STF.translate_upper_case(self.sequence, self.frame, condensed=False)
            string_2 = STF.translate_upper_case(self.mutated_sequence, self.frame, condensed=False)
        elif type == 'DNA-Final':
            string_1 = self.sequence
            string_2 = self.mutated_sequence
            if self.gRNA.cutsite > self.insert_site:
                mutation_range = range(self.insert_site, self.gRNA.cutsite)
            else:
                mutation_range = range(self.gRNA.cutsite, self.insert_site)
            if self.base_translation != self.mutant_translation:
                if self.large_deletion_range == None:
                    incorporation_warning = True
            if self.tag_type in ['I', 'S']:
                incorporation_warning = True
        else:
            return ''
        length = len(string_1)
        if len(string_2) < len(string_1):
            length = len(string_2)
        output = ''
        for i in range(length):
            if string_1[i] == string_2[i]:
                output += ' '
            else:
                if type == 'DNA-Final':
                    if incorporation_warning == True:
                        if i in mutation_range:
                            output += '*'
                        else:
                            output += '~'
                    else:
                        output += '*'
                else:
                    output += '*'
        if len(string_2) < len(string_1):
            for i in range(len(string_1) - len(string_2)):
                output += '!'

        return output

    def feature_locator(self, feature):
        """ input is a keyword string: insert_site, guide, or mutation. Returns string '1.x', x is location
            of referenced feature. This is useful for using the See function to scroll to the feature after
            display updates. """
        output = '1.'
        if feature == 'insert_site':
            output += str(self.insert_site)
        if feature == 'guide':
            output += str(self.gRNA.cutsite)
        if feature == 'mutation':
            output += str(self.mutation_site)

        return output

    def guess_frame(self):
        """ Sets the initial translation frame so that insertion site is in-frame
            Should always be correct for in-coding region inserts, can guess wrong for intron / syntron tags"""
        coding_base_count = 0
        for base in self.sequence[: self.insert_site]:
            if base.isupper():
                coding_base_count += 1
        frame = coding_base_count % 3
        return frame

    def next_base_ID(self, sequence, number):
        count = 0
        output = ''
        while len(output) < number:
            try:
                if sequence[count] in 'AGTCagtc':
                    output += sequence[count]
                count += 1
            except IndexError:
                output += '--'

        return output

    def set_pointer(self, position, marker='|'):
        """ Generates pointer string for display. Sets self.pointer_string, and returns string. """
        output = ''
        for i in range(len(self.sequence)):
            if i == position:
                output += marker
            else:
                output += ' '
        self.pointer_string = output

        return output

    def incorporate_mutation(self, position, mutation):
        """ Simply changes the base at position in sequence to mutation within self.mutated_sequence attribute. """
        self.mutated_sequence = DNA(self.mutated_sequence[ : position] + mutation +
                                    self.mutated_sequence[position + 1: ])
        return 0

    def change_frame(self, increment):
        """ Input increment (1 or -1) sets self.frame to 0, 1, or 2 based on increment plus set value. """
        new_frame = (self.frame + increment) % 3
        self.frame = new_frame
        return new_frame

    def get_current_codon(self, position):
        """ Input = position in self.mutated_sequence. Locates the codon the position is part of.
            Returns a library with the positions of the entire codon, the codon, and the amino acid. """
        upper_count = 0
        for base in self.mutated_sequence[:position + 1]:
            if base.isupper():
                upper_count += 1
        position_in_codon = (upper_count + 2 - self.frame) % 3
        # Initialize position list to locate positions of 2 coding bases upstream and 2 downstream of active position
        codon_positions = ['', '', '', '', '']
        codon_positions[2] = position
        # Walk downstream of position, record locations of next 2 uppercase bases in slots 3 and 4 of position list
        count = 0
        test_position = position
        while count < 2:
            test_position += 1
            try:
                if self.mutated_sequence[test_position].isupper():
                    count += 1
                    codon_positions[2 + count] = test_position
            except IndexError:
                count = 2
            if test_position > len(self.mutated_sequence):
                count = 2
        # Walk downstream of position, record locations of next 2 uppercase bases in slots 0 and 1 of position list
        count = 0
        test_position = position
        while count > -2:
            test_position -= 1
            try:
                if self.mutated_sequence[test_position].isupper():
                    count -= 1
                    codon_positions[2 + count] = test_position
            except IndexError:
                count = -2
            if test_position <= 0:
                count = -2
        # Remove the extra positions from the position list depending on the location of original position in the codon
        codon_trimmer = [[0, 0], [0, 3], [3, 3]]
        for item in codon_trimmer[position_in_codon]:
            codon_positions.pop(item)
        # Return an empty set if no complete codon could be found within provided sequence
        for i in range(0, 3):
            if codon_positions[i] == '':
                return {}
        # include codon in return
        codon = ''
        for i in codon_positions:
            codon += self.mutated_sequence[i]

        amino_acid = STF.translate(codon)
        # include amino acid in return
        output_library = {'positions_list': codon_positions, 'codon': codon, 'amino_acid': amino_acid}

        return output_library

    def generate_SapI_string(self):
        """ refers to self.mutated_string attribute. Builds string that displays only white-space and sapI sites
            This shows user clearly where SapI sites are in sequence, and clears them when mutated.
            SapI_flag attribute may not be used by current version of program. """
        # Trim deletions from reference string and convert to lower case
        mutated_sequence = self.mutated_sequence.lower()
        reference_string = ''
        count = 0
        deleted_base_positions = []
        while count < len(mutated_sequence):
            base = mutated_sequence[count]
            if base in 'acgt':
                reference_string += base
            elif base == '-':
                deleted_base_positions.append(count)
            count += 1
        # Find sapI sites in reference string and store the locations of the bases
        sapI_site_start_positions = []
        for site in ['gctcttc', 'gaagagc']:
            start = 0
            sapI_site_start = reference_string.find(site, start)
            while sapI_site_start != -1:
                sapI_site_start_positions.append(sapI_site_start)
                start = sapI_site_start + 1
                sapI_site_start = reference_string.find(site, start)
        # From sapI_site_start_positions and deleted base locations, get base coordinates for all bases in sapI sites
        output = ''
        if sapI_site_start_positions != []:
            output_locations = []
            for position in sapI_site_start_positions:
                for i in range(7):
                    output_locations.append(position + i)
            for i in range(len(reference_string)):
                if i in output_locations:
                    output += reference_string[i]
                else:
                    output += ' '
            if deleted_base_positions != []:
                for position in deleted_base_positions:
                    char = ' '
                    if output[position - 1] != ' ':
                        if output[position] != ' ':
                            char = '-'
                    output = output[:position] + char + output[position:]
        self.SapI_string = output
        self.sapI_flag = len(sapI_site_start_positions)
        return len(sapI_site_start_positions)

    def review_flags(self):
        """ Inspected gRNA binding site. If in self.mutated_sequence, gRNA binding site is not disrupted by either
            insertion or mutation, prints a flag to alert user.
            Also appends any homology Arm flags. Outputs string to print on review display.
            This is a huge function with complicated flow. Consider simplifying by adding a disruption mask
            to the gRNA specification class."""
        # store variables locally
        gRNA_length = len(self.gRNA.sequence)
        binding_site_length = len(self.gRNA.binding_site)
        gRNA_start = self.gRNA.start
        mutated_sequence = str(self.mutated_sequence)
        guide_specification = self.gRNA.guide_specification
        gRNA_sequence = str(self.gRNA.sequence)
        insert_location = self.insert_site - self.gRNA.start
        if self.gRNA.direction == 'antisense':
            gRNA_start = len(mutated_sequence) - self.gRNA.start
            mutated_sequence = self.reverse_complement(mutated_sequence)
            insert_location = len(mutated_sequence) - self.insert_site - gRNA_start
        if insert_location == 0:
            insert_location -= 1
        PAM_locations = []
        seed_locations = []
        weak_locations = []
        PAM_spacer_locations = []
        PAM_location_offset = 0
        binding_site_offset = 0
        if guide_specification.PAM_location ==  'RIGHT':
            PAM_location_offset -= guide_specification.bind_start
        elif guide_specification.PAM_location == 'LEFT':
            binding_site_offset += len(guide_specification.PAM)
        for i in range(len(guide_specification.PAM)):
            if guide_specification.PAM[i].lower() != 'n':
                PAM_locations.append(PAM_location_offset + i)
            else:
                PAM_spacer_locations.append(PAM_location_offset + i)
        for i in range(binding_site_offset, binding_site_offset + binding_site_length):
            if guide_specification.PAM_location == 'RIGHT':
                if binding_site_length - i < 9:
                    seed_locations.append(i)
                else:
                    weak_locations.append(i)
            elif guide_specification.PAM_location == 'LEFT':
                if i < 8 + len(guide_specification.PAM):
                    seed_locations.append(i)
                else:
                    weak_locations.append(i)
    # Get gRNA binding site from repair template
        for position in range(len(PAM_locations)):
            while mutated_sequence[gRNA_start + position] == '-':
                if gRNA_start > 0:
                    gRNA_start -= 1
                else:
                    break
        sequence_length_count = 0
        gRNA_binding_site = ''
        while sequence_length_count < gRNA_length:
            if mutated_sequence[gRNA_start + sequence_length_count] == '-':
                gRNA_start += 1
            else:
                gRNA_binding_site += mutated_sequence[gRNA_start + sequence_length_count]
                sequence_length_count += 1
    # Check repair template binding site vs WT binding site for differences
        flags = []
        variant_locations = []
        output = ''
        spacer = ''
        for i in range(self.insert_site - 25):
            spacer += ' '
        for position in range(gRNA_length):
            if gRNA_binding_site[position] != gRNA_sequence[position]:
                variant_locations.append(position)
        mutation_categories = [PAM_locations, seed_locations, weak_locations]
        for category in mutation_categories:
            for location in category:
                if location in variant_locations:
                    if category in mutation_categories:
                        mutation_categories.remove(category)
        if PAM_locations in mutation_categories:
            PAM_locations += PAM_spacer_locations
        if seed_locations in mutation_categories:
            if max(seed_locations) > max(weak_locations):
                seed_locations += weak_locations[-7:]
            else:
                seed_locations += weak_locations[:7]
        for category in mutation_categories:
            if insert_location in category:
                if category in mutation_categories:
                    mutation_categories.remove(category)
        if PAM_locations in mutation_categories:
            if seed_locations in mutation_categories:
                if weak_locations in mutation_categories:
                    flags.append('gRNA binding site intact. Repair template can be cut by Cas9 + guideRNA.')
                else:
                    flags.append('gRNA binding weakly disrupted. Consider mutations nearer to PAM.')
        for flag in flags:
            if flag not in self.homology_arm_flags:
                self.homology_arm_flags.append(flag)
        for flag in self.homology_arm_flags:
            output += spacer + flag

        return output

    def homology_arm_input(self, type):
        """ Input is type, 'hUS' or 'hDS'
            Uses variables stored in attributes to extract information used by the homology_arm_generator
            Returns libary of information that feeds into homology_arm generator.
            Must be called independently for hUS and hDS
            Strips ATG from N-tags
            Determines if cutsite lies within arm
            Cuts sequence in half at the cutsite, passes only relevant half in output"""
        # Store object attributes locally for easier reading code
        selector_sequence = self.mutated_sequence
        insert_site = self.insert_site
        cutsite = self.gRNA.cutsite
        # Strip ATG from sequence if N-tag (reappended as sticky end)
        if self.tag_type == 'N':
            if selector_sequence[insert_site - 3: insert_site] == 'ATG':
                selector_sequence = selector_sequence[:insert_site - 3] + '---' + selector_sequence[insert_site:]
            elif selector_sequence[insert_site:insert_site + 3] == 'ATG':
                selector_sequence = selector_sequence[:insert_site] + '---' + selector_sequence[insert_site + 3:]
        if type == 'hUS':
            if cutsite >= insert_site:
                cutsite = None
            else:
                cutsite = insert_site - cutsite
            selector_sequence = self.reverse_complement(selector_sequence[: insert_site])
            target_length = self.hUS_target_length
        elif type == 'hDS':
            if cutsite <= insert_site:
                cutsite = None
            else:
                cutsite = cutsite - insert_site
            selector_sequence = selector_sequence[insert_site: ]
            target_length = self.hDS_target_length
        if self.tag_type == 'S':
            target_length -= 3
        # hUS is passed as reverse complement. homology_arm_generator counts outward from insert_site to get
        # full length homology arm, passing one as reverse complement allows one function to handle both
        return {'sequence':selector_sequence, 'cutsite':cutsite, 'target_length':target_length, 'type':type}

    def homology_arm_generator(self, HA_library):
        """ Input is library of information from homology_arm_input method.
            Output is library with 2 elements: 'HA' = homology arm (deletions removed) for final reagent construction
            'TS' is precursor to HA that contains deletions. Used for alignment / display
            Adds flags to self.homolgy_arm_flags if lengths are altered from target. """
        sequence = HA_library['sequence']
        cutsite = HA_library['cutsite']
        target_length = HA_library['target_length']

        minimum_true_arm = self.homology_arm_minimum_length
        homology_arm_sequence = ''

        if cutsite != None:
            if self.large_deletion_range != None:   # If cutsite in large deletion, can be distant from "insert" site
                if self.gRNA.cutsite in range(int(self.large_deletion_range[0]), int(self.large_deletion_range[1])):
                    cutsite = 0
            if cutsite > target_length - minimum_true_arm:
                target_length = cutsite + minimum_true_arm
                flag = '{0} homology arm extended beyond target length because ' \
                                           'cutsite within minimum distance of edge.'.format(HA_library['type'])
                if flag not in self.homology_arm_flags:
                    self.homology_arm_flags.append(flag)

        elif cutsite == None:
            if minimum_true_arm > target_length:
                target_length = minimum_true_arm
                flag = '{0} homology arm extended beyond target length because ' \
                                           'minimum arm length > target length.'.format(HA_library['type'])
                if flag not in self.homology_arm_flags:
                    self.homology_arm_flags.append(flag)

        if target_length > len(sequence) - sequence.count('-'):
            flag = 'Insufficient sequence provided for desired length ' \
                                       '{0} homology arm.'.format(HA_library['type'])
            if flag not in self.homology_arm_flags:
                self.homology_arm_flags.append(flag)
        truncation_site = target_length
        continue_loop = True
        while continue_loop == True:
            if truncation_site >= len(sequence):
                if len(sequence) - sequence.count('-') < target_length:
                    flag = 'Insufficient sequence provided for desired length ' \
                                               '{0} homology arm.'.format(HA_library['type'])
                    if flag not in self.homology_arm_flags:
                        self.homology_arm_flags.append(flag)
                truncation_site = len(sequence)
                continue_loop = False
            elif len(sequence[:truncation_site]) - sequence[:truncation_site].count('-') < target_length:
                truncation_site += 1
            elif len(sequence[:truncation_site]) - sequence[:truncation_site].count('-') == target_length:
                continue_loop = False

        trimmed_sequence = sequence[:truncation_site]
        for base in trimmed_sequence:
            if base in 'ACGTacgt':
                homology_arm_sequence += base
        if HA_library['type'] == 'hUS':
            homology_arm_sequence = self.reverse_complement(homology_arm_sequence)
            trimmed_sequence = self.reverse_complement(trimmed_sequence)

        return {'HA':homology_arm_sequence, 'TS':trimmed_sequence}

    def get_homology_arms(self):
        """ Controller for homology arm input and generator functions.
            When called, calls functions in succession.
            Function sets output to self.hUS and self.hDS attributes. """
        self.hUS = self.homology_arm_generator(self.homology_arm_input('hUS'))['HA']
        self.hDS = self.homology_arm_generator(self.homology_arm_input('hDS'))['HA']

        return 0

    def site_review(self):
        """ Gets raw homology arm sequences by running homology_arm input/generator
            Sets self.repiar_template string for use in review display
            appends homolgy arm flags for sapI sites within homology arm. """
        if self.SapI_string is None:
            self.generate_SapI_string()
        hUS = self.homology_arm_generator(self.homology_arm_input('hUS'))
        hDS = self.homology_arm_generator(self.homology_arm_input('hDS'))


        hUS_start = self.insert_site - len(hUS['TS'])
        hDS_end = self.insert_site + len(hDS['TS'])
        flags = []
        spacer = ''
        for i in range(hUS_start):
            spacer += ' '
        repair_string = spacer + hUS['TS'] + hDS['TS']
        #repair_string = self.flush_string(repair_string)
        self.repair_template = repair_string
        if self.sap_option == 'SapTrap':
            for site in ['gctcttc', 'gaagagc']:
                for arm in [hUS['HA'], hDS['HA']]:
                    if site in arm.lower():
                        flags.append('SapI site in repair template. Incompatible with SapTrap assembly.')
        if '~' in self.DNA_difference_string:
            flags.append('For point mutagenesis only: Mutations marked by ~ have a lower probability of incorporation in genome. Position critical mutations between the cutsite and the insert site.')

        for flag in flags:
            if flag not in self.homology_arm_flags:
                self.homology_arm_flags.append(flag)
        return repair_string

    def trim_leading_spaces(self, sequence):

        while sequence[0] == ' ':
            sequence = sequence[1:]
        return sequence

    def print_oligos(self, view):
        """ Input is string specifying the desired format for the output.
            Returns block of text containing reagent sequences in desired format for user"""

        output = ''
        flag_count = 0
        for flagset in [self.homology_arm_flags, self.guide_flags]:
            if flagset != []:
                for flag in flagset:
                    output += self.trim_leading_spaces(flag) + '\n'
                    flag_count += 1
        output += '\n'

        if self.sap_option == 'Generic':
            output += 'Homology arm and guide RNA sequences:\n'
            output += 'hUS:   ' + self.hUS + '\n' + \
                      'hDS:   ' + self.hDS + '\n' + \
                      'sgRNA: ' + self.gRNA.binding_site + '\n'

        elif self.sap_option == 'SapTrap':
            tag_type = self.tag_type
            hUS_seq = self.hUS
            hDS_seq = self.hDS
            gRNA_seq = self.gRNA.binding_site
            sticky_ends = {'N': ['TTG', 'AAC', 'TGG', 'CAT', 'ACG', 'TAC'],
                           'C': ['TTG', 'AAC', 'TGG', 'CGC', 'GGT', 'TAC'],
                           'L': ['TTG', 'AAC', 'TGG', 'CGC', 'ACG', 'TAC'],
                           'X': ['TTG', 'AAC', 'TGG', 'CAT', 'GGT', 'TAC'],
                           'I': ['TTG', 'AAC', 'TGG', 'act', 'ttt', 'TAC'],
                           'S': ['TTG', 'AAC', 'TGG', 'acttac', 'tttcag', 'TAC']}
            hUS_T = sticky_ends[tag_type][2] + hUS_seq
            hUS_B = sticky_ends[tag_type][3] + self.reverse_complement(hUS_seq)
            hDS_T = sticky_ends[tag_type][4] + hDS_seq
            hDS_B = sticky_ends[tag_type][5] + self.reverse_complement(hDS_seq)
            sgRNA_T = sticky_ends[tag_type][0] + gRNA_seq
            sgRNA_B = sticky_ends[tag_type][1] + self.reverse_complement(gRNA_seq)
            if tag_type == 'S':
                hUS_T += 'gta'
                hDS_B += 'ctg'

            if view == 'Alignment':
                output += 'Alignment of annealed oligos:\n'
                output += hUS_T + '\n'
                output += '   ' + self.reverse(hUS_B) + '\n\n'
                output += hDS_T + '\n'
                output += '   ' + self.reverse(hDS_B) + '\n\n'
                output += sgRNA_T + '\n'
                output += '   ' + self.reverse(sgRNA_B) + '\n\n'

            if 'Oligos' in view:
                output += "Oligos in 5' to 3' direction\n\n"
                output += self.jobname + ' hUS_T\t' + hUS_T + '\n'
                output += self.jobname + ' hUS_B\t' + hUS_B + '\n'
                output += self.jobname + ' hDS_T\t' + hDS_T + '\n'
                output += self.jobname + ' hDS_B\t' + hDS_B + '\n'
                output += self.jobname + ' sgRNA_T\t' + sgRNA_T + '\n'
                output += self.jobname + ' sgRNA_B\t' + sgRNA_B + '\n\n'

            synthetic_DNA = 'gtgGCTCTTCg' + sticky_ends[tag_type][0] + gRNA_seq + self.reverse_complement(
                sticky_ends[tag_type][1]) + \
                            'cGAAGAGC' + \
                            'gtgGCTCTTCg' + sticky_ends[tag_type][2] + hUS_seq + self.reverse_complement(
                sticky_ends[tag_type][3]) + \
                            'cGAAGAGC' + \
                            'gtgGCTCTTCg' + sticky_ends[tag_type][4] + hDS_seq + self.reverse_complement(
                sticky_ends[tag_type][5]) + \
                            'cGAAGAGCgtg\n\n'

            if 'Synthetic' in view:
                if 'Annotated' not in view:
                    output += 'Synthetic DNA\n\n'
                    output += synthetic_DNA

            if 'Annotated' in view:

                gRNA_len = len(self.gRNA.binding_site)
                hUS_len = len(self.hUS)
                hDS_len = len(self.hDS)
                if self.tag_type == 'S':
                    hUS_len += 3
                    hDS_len += 3
                total_length = 78 + gRNA_len + hUS_len + hDS_len

                output += 'Annotated Synthetic DNA:\n\n'

                list_annotation = [['spacer', 3], ['SapI', 7], ['spacer', 1], ['sticky', 3], ['gRNA', gRNA_len],
                                   ['sticky', 3], ['spacer', 1], ['SapI', 7], ['spacer', 3], ['SapI', 7],
                                   ['spacer', 1], ['sticky', 3], ['hUS', hUS_len], ['sticky', 3], ['spacer', 1],
                                   ['SapI', 7], ['spacer', 3], ['SapI', 7], ['spacer', 1], ['sticky', 3],
                                   ['hDS', hDS_len], ['sticky', 3], ['spacer', 1], ['SapI', 7]]

                count=1
                sapI_counter = 0
                sticky_counter = 0
                list_output_constructor = []
                for feature in list_annotation:
                    feature_location = str(count) + '..' + str(count + feature[1] - 1)
                    count += feature[1]
                    if feature[0] == 'spacer':
                        continue
                    elif feature[0] == 'SapI':
                        feature_label = 'SapI_site'
                        if sapI_counter > 0:
                            feature_label += '(' + str(sapI_counter) + ')'
                        feature_color = "#ff5375"
                        sapI_counter += 1
                    elif feature[0] == 'sticky':
                        feature_label = 'SapTrap overhang'
                        if sticky_counter > 0:
                            feature_label += '(' + str(sticky_counter) + ')'
                        feature_color = 'cyan'
                        sticky_counter += 1
                    elif feature[0] == 'gRNA':
                        feature_label = self.jobname + ' sgRNA'
                        feature_color = "#ff974a"
                    elif feature[0] == 'hUS':
                        feature_label = self.jobname + ' hUS'
                        feature_color = "#68a368"
                    elif feature[0] == 'hDS':
                        feature_label = self.jobname + ' hDS'
                        feature_color = "#68a368"
                    list_output_constructor.append([feature_location, feature_label, feature_color])


                annotated_file = 'LOCUS\t\t' + self.jobname + '\t\t\t' + str(total_length) + \
                             ' bp\nCOMMENT     ApEinfo: methylated:1\nFEATURES             ' \
                             'Location/Qualifiers\n'
                for item in list_output_constructor:
                    annotated_file += '     misc_feature    ' + item[0] + '\n                     /locus_tag="' + item[1] +   \
                    '"\n                     /label="' + item[1] + '"\n                     /ApEinfo_label="' + item[1] +        \
                    '"\n                     /ApEinfo_fwdcolor="' + item[2] + '"\n                     /ApEinfo_revcolor="' +    \
                    item[2] + '"\n                     /ApEinfo_graphicformat="arrow_data {{0 1 2 0 0 -1} {} 0}' +    \
                    '\n                     width 5 offset 0"\n'
                annotated_file += 'ORIGIN\n' + synthetic_DNA + '//'
                output += annotated_file

        return {'labels' : '', 'sequences' : output, 'flags' : list(range(flag_count + 1))}

class DisplayStrings:
    """ Have this start with a single library: keys are ordered display line items and values are lists that
    include labels, pointers to data, etc.
    this should make it easy to add display lines later
    """
    def __init__(self, edit_site):
        self.edit_site = edit_site
        # Need to make sure these labels match entries in fetch_display_strings
        self.string_labels = ['base_translation', 'translation_differences', 'mutant_translation', 'pointer',
                              'mutated_sequence', 'repair_template', 'DNA_differences', 'sequence',
                              'gRNA', 'PAM', 'SapI_sites', 'homology_arm_flags']
        self.active_string_labels = []
        self.active_strings = []
        self.output_lines = None

    def reset(self, edit_site):
        self.edit_site = edit_site
        self.active_string_labels = []
        self.active_strings = []
        self.output_lines = None

    def add_display_strings(self, *args):
        """ input = string that corresponds to [0] position of directory list entry for display string to append
            output appends the label to the active_strings attribute """
        for argument in args:
            if argument in self.active_string_labels:
                pass
            elif argument in self.string_labels:
                self.active_string_labels.append(argument)

        return 0

    def remove_display_strings(self, *args):
        """ checks if argument string is in active strings, removes if present """
        for argument in args:
            if argument == 'repair_template':
                if argument in self.active_string_labels:
                    self.add_display_strings('mutated_sequence')
            if argument in self.active_string_labels:
                self.active_string_labels.remove(argument)

        return 0
    def flush_string(self, string):
        """ Input = string. compares length to self.sequence and adds spaces to input to make strings same length.
            Returns original string with spaces appended. """
        while len(string) < len(self.edit_site.sequence):
            string += ' '
        return string

    def fetch_display_strings(self):
        """ fetches full string directory entries from string directory for those strings represented in
            active_string_labels.
            returns an ordered sub directory in same style as display_string_directory for only active strings"""

        ordered_display_strings = []
        if 'base_translation' in self.active_string_labels:
            ordered_display_strings.append(['base_translation', 'Wild Type Translation:',
                                            self.edit_site.base_translation])
        if 'translation_differences' in self.active_string_labels:
            ordered_display_strings.append(['translation_differences', '',
                                            self.edit_site.translation_difference_string])
        if 'mutant_translation' in self.active_string_labels:
            ordered_display_strings.append(['mutant_translation', 'Repair Template Translation:',
                                            self.edit_site.mutant_translation])
        if 'pointer' in self.active_string_labels:
            ordered_display_strings.append(['pointer', '', self.edit_site.pointer_string])
        if 'mutated_sequence' in self.active_string_labels:
            ordered_display_strings.append(['mutated_sequence', 'Repair Template DNA:',
                                            str(self.edit_site.mutated_sequence)])
        if 'repair_template' in self.active_string_labels:
            ordered_display_strings.append(['repair_template', 'Repair Template:', str(self.edit_site.repair_template)])
        if 'DNA_differences' in self.active_string_labels:
            ordered_display_strings.append(['DNA_differences', '', self.edit_site.DNA_difference_string])
        if 'sequence' in self.active_string_labels:
            ordered_display_strings.append(['sequence', 'Wild Type DNA:', str(self.edit_site.sequence)])
        if 'gRNA' in self.active_string_labels:
            if self.edit_site.gRNA is None:
                pass
            else:
                ordered_display_strings.append(['gRNA', 'Guide RNA binding site:',
                                                self.edit_site.gRNA_string])
        if 'PAM' in self.active_string_labels:
            if self.edit_site.gRNA is None:
                pass
            else:
                ordered_display_strings.append(['PAM', 'PAM:', self.edit_site.PAM_string])
            if self.edit_site.guide_flags is None:
                pass
            else:
                for item in self.edit_site.guide_flags:
                    ordered_display_strings.append(['flag', '', item])
        if 'SapI_sites' in self.active_string_labels:
            if self.edit_site.SapI_string is None:
                self.edit_site.generate_SapI_string()
            ordered_display_strings.append(['SapI_sites', 'SapI Sites:', self.edit_site.SapI_string])
        if 'homology_arm_flags' in self.active_string_labels:
            if self.edit_site.homology_arm_flags is None:
                pass
            else:
                flag_pad_length = 0
                if self.edit_site.gRNA is None:
                    flag_pad_length = self.edit_site.insert_site - 30
                else:
                    count = 0
                    while self.edit_site.gRNA_string[count] == ' ':
                        flag_pad_length += 1
                        count += 1
                count = 0
                flag_pad = ''
                while count < flag_pad_length:
                    flag_pad += ' '
                    count += 1
                for item in self.edit_site.homology_arm_flags:
                    ordered_display_strings.append(['flag', '', flag_pad + item])

        i = 0
        while i < len(ordered_display_strings):
            ordered_display_strings[i][2] = self.flush_string(ordered_display_strings[i][2])
            i += 1
        return ordered_display_strings

    def format_display_strings(self):
        labels = ''
        sequences = ''
        flags = []
        ins = self.edit_site.insert_site
        display_insert_strings = ['mutated_sequence', 'repair_template', 'sequence', 'gRNA']
        ordered_display_strings = self.fetch_display_strings()
        i = 0
        for item in ordered_display_strings:
            i += 1
            if item[0] == 'flag':
                flags.append(i)
        if self.edit_site.show_insert_site == True:
            for item in ordered_display_strings:
                if item[0] in display_insert_strings:
                    if self.edit_site.guess_frame() == self.edit_site.frame:
                        insert_text = '|-INSERT-|'
                    else:
                        insert_text = '|-insert-|'
                    item[2] = item[2][:ins] + insert_text + item[2][ins:]
                elif item[0] == 'flag':
                    count = 0
                    while item[2][count] == ' ':
                        count += 1
                    if self.edit_site.insert_site <= count:
                        item[2] = item[2][:ins] + '          ' + item[2][ins:]
                else:
                    item[2] = item[2][:ins] + '          ' + item[2][ins:]

        for item in ordered_display_strings:
            labels += item[1]
            labels += '\n'
            sequences += item[2]
            sequences += '\n'

        self.output_lines = sequences.count('\n')

        return {'labels' : labels, 'sequences' : sequences, 'flags' : flags}

