# Copyright (C) 2018 Matthew L. Schwartz - all rights reserved
# Not for redistribution


def format_sequence(sequence):
    """Input = user supplied string containing DNA sequence, \n's, and a single non-DNA character
       Output is a library with two items: 'sequence' is the DNA sequence, 'insertion_site' is the
       position in the sequence marked for insertion
       if more than one insertion site is marked, sequence will be empty string"""

    if sequence == 'CAT':
        sequence = ''
    clean_sequence = ''
    insertion_site = int(len(sequence)/2)                   #default value is middle of input sequence
    non_base_char = 0                       # insertion site is set to the midpoint by default, incase indicator omitted
    sequence_summary = {'deletion' : False, 'change_nuclease' : False}
    del_region = [0, 0]
    nonbase_characters = []
    for char in sequence:                           # Loop will remove any characters that are not AGCTagct
        if char in 'AGCTagct':
            clean_sequence += char
        # Attempt to add deletion support
        elif char == '<':
            del_region[0] = int(len(clean_sequence))
            insertion_site = del_region[0] + 1
            nonbase_characters.append(char)
            continue
        elif char == '>':
            del_region[1] = int(len(clean_sequence))
            sequence_summary['deletion'] = True
            sequence_summary['deletion_region'] = del_region
            nonbase_characters.append(char)
            continue
        # End attempt to add deletion support
        elif char == '\n':
            continue
        elif char == '/':                                           # records insertion site as position of non-AGCTagct character
            insertion_site = int(len(clean_sequence))
            nonbase_characters.append(char)
        elif char == '#':
            sequence_summary['change_nuclease'] = True
            continue
        else:
            nonbase_characters.append(char)
    if nonbase_characters not in [[], ['<', '>'], ['/']]:                               # re-prompt if 1 non-base characters were entered
        clean_sequence = 'Entry_Error'
    sequence_summary['sequence'] = clean_sequence
    sequence_summary['insertion_site'] = insertion_site
    return sequence_summary

def degenerate_sequence_generator(sequence, output_case='neutral'):
    """ Input is a DNA sequence object with degenerate bases
        Output is a list of all possible permutations of the sequence """
    bases = 'ACGTacgt'
    degenerate_code = {'B': ['C', 'G', 'T'], 'D': ['A', 'G', 'T'], 'H': ['A', 'C', 'T'], 'V': ['A', 'C', 'G'],
                       'K': ['G', 'T'], 'M': ['A', 'C'], 'R': ['A', 'G'], 'S': ['C', 'G'], 'W': ['A', 'T'],
                       'Y': ['C', 'T'], 'N': ['A', 'C', 'G', 'T']}

    possible_sequences = ['']
    for base in sequence:
        if base.upper() in degenerate_code.keys():
            number_of_new_possibilities = len(possible_sequences) * \
                                          len(degenerate_code[base.upper()]) - len(possible_sequences)
            count = 0
            while count < number_of_new_possibilities:
                possible_sequences.append('')
                count += 1
    count = 0
    sequences_per_base = len(possible_sequences)
    while count < len(sequence):
        base = sequence[count]
        permutation_count = 0
        if base in bases:
            while permutation_count < len(possible_sequences):
                possible_sequences[permutation_count] += base
                permutation_count += 1
        elif base.upper() in degenerate_code.keys():
            number_of_possible_bases = len(degenerate_code[base.upper()])
            sequences_per_base /= number_of_possible_bases
            base_array = []
            for possible_base in degenerate_code[base.upper()]:
                base_counter = 0
                while base_counter < sequences_per_base:
                    base_array.append(possible_base)
                    base_counter += 1
            substitution_array = []
            while len(substitution_array) < len(possible_sequences):
                substitution_array += base_array
            while permutation_count < len(possible_sequences):
                if base.islower():
                    base_to_add = substitution_array[permutation_count].lower()
                else:
                    base_to_add = substitution_array[permutation_count]
                possible_sequences[permutation_count] += base_to_add
                permutation_count += 1
        count += 1

    if output_case == 'neutral':
        return possible_sequences
    elif output_case == 'upper':
        return_sequences = []
        for sequence in possible_sequences:
            return_sequences.append(sequence.upper())
        return return_sequences
    elif output_case == 'lower':
        return_sequences = []
        for sequence in possible_sequences:
            return_sequences.append(sequence.lower())
        return return_sequences


def list_to_lowercase(list):
    """ Input a list of strings // Output a list of strings converted to lowercase"""
    lower_case_list = []
    for item in list:
        lower_case_list.append(item.lower())
    return lower_case_list




dictGeneticCode = {     'TTT' : 'F', 'TTC' : 'F', 'TTA' : 'L', 'TTG' : 'L',
                        'CTT' : 'L', 'CTC' : 'L', 'CTA' : 'L', 'CTG' : 'L',
                        'ATT' : 'I', 'ATC' : 'I', 'ATA' : 'I', 'ATG' : 'M',
                        'GTT' : 'V', 'GTC' : 'V', 'GTA' : 'V', 'GTG' : 'V',
                        'TCT' : 'S', 'TCC' : 'S', 'TCA' : 'S', 'TCG' : 'S',
                        'CCT' : 'P', 'CCC' : 'P', 'CCA' : 'P', 'CCG' : 'P',
                        'ACT' : 'T', 'ACC' : 'T', 'ACA' : 'T', 'ACG' : 'T',
                        'GCT' : 'A', 'GCC' : 'A', 'GCA' : 'A', 'GCG' : 'A',
                        'TAT' : 'Y', 'TAC' : 'Y', 'TAA' : '*', 'TAG' : '*',
                        'CAT' : 'H', 'CAC' : 'H', 'CAA' : 'Q', 'CAG' : 'Q',
                        'AAT' : 'N', 'AAC' : 'N', 'AAA' : 'K', 'AAG' : 'K',
                        'GAT' : 'D', 'GAC' : 'D', 'GAA' : 'E', 'GAG' : 'E',
                        'TGT' : 'C', 'TGC' : 'C', 'TGA' : '*', 'TGG' : 'W',
                        'CGT' : 'R', 'CGC' : 'R', 'CGA' : 'R', 'CGG' : 'R',
                        'AGT' : 'S', 'AGC' : 'S', 'AGA' : 'R', 'AGG' : 'R',
                        'GGT' : 'G', 'GGC' : 'G', 'GGA' : 'G', 'GGG' : 'G',
                       }

dict_rev_genetic_code = {'A' : ['GCT', 'GCA', 'GCC', 'GCG'],
                         'C' : ['TGT', 'TGC' ],
                         'D' : ['GAT', 'GAC' ],
                         'E' : ['GAA', 'GAG' ],
                         'F' : ['TTT', 'TTC' ],
                         'G' : ['GGA', 'GGT', 'GGC', 'GGG' ],
                         'H' : ['CAT', 'CAC' ],
                         'I' : ['ATT', 'ATC', 'ATA'],
                         'K' : ['AAA', 'AAG' ],
                         'L' : ['CTT', 'TTG', 'CTC', 'CTG', 'TTA', 'CTA'],
                         'M' : ['ATG'],
                         'N' : ['AAT', 'AAC' ],
                         'P' : ['CCA', 'CCG', 'CCT', 'CCC'],
                         'Q' : ['CAA', 'CAG' ],
                         'R' : ['AGA', 'CGA', 'CGT', 'CGC', 'CGG', 'AGG' ],
                         'S' : ['TCA', 'TCT', 'TCG', 'AGT', 'TCC', 'AGC' ],
                         'T' : ['ACA', 'ACT', 'ACC', 'ACG' ],
                         'V' : ['GTT', 'GTG', 'GTC', 'GTA' ],
                         'W' : ['TGG' ],
                         'Y' : ['TAT', 'TAC' ],
                         '*' : ['TAA', 'TGA', 'TAG' ] }

def synonymous_codons(amino_acid):
    """ Argument is single letter amino acid code
        Returns list of codons for amino acid """
    try:
        codon_options = dict_rev_genetic_code[amino_acid.upper()]
    except KeyError:
        codon_options = []
    return codon_options

def translate(sequence):
    """ Input DNA sequence of upper-case characters. Output translated protein sequence."""
    translated_sequence = ''
    for base in sequence:
        if base.upper() in 'ACGT':
            translated_sequence += base.upper()
    if len(translated_sequence) % 3 == 0:                                      # if test trims sequence to multiple of 3
        coding_length = len(translated_sequence)
    else:
        coding_length = len(translated_sequence) - (len(translated_sequence) % 3)
    translation = ''                                                 # initialize storage string and counter
    count = 0
    while count < coding_length:                                    # loops through codons, appends to translated string

        codon = translated_sequence[count : count + 3]
        translation += dictGeneticCode[codon]
        count += 3
    return translation

def translate_upper_case(seq, frame, condensed=True):
    """ Input sequence of DNA characters and translation frame. Uppercase will be translated, lowercase discarded.
        Output translated amino acids. condensed leaves no spaces between amino acids, False spaces for DNA alignment"""
    parsed_seq = []
    seq_run = ''
    seq_translate = ''
    if seq[0].islower():
        prev_base_case = 'lower'
    elif seq[0] == '-':
        prev_base_case = 'lower'
    elif seq[0].isupper():
        prev_base_case = 'upper'
    for base in seq:
        if base.islower():
            base_case = 'lower'
        elif base == '-':
            base_case = 'lower'
        elif base.isupper():
            base_case = 'upper'
        if base_case == prev_base_case:
            seq_run = seq_run + base
        else:
            parsed_seq.append(seq_run)
            seq_run = '' + base
            if base.islower():
                prev_base_case = 'lower'
            elif base.isupper():
                prev_base_case = 'upper'
    parsed_seq.append(seq_run)

    for item in parsed_seq:
        if item.isupper():
            seq_translate += item
    protein_seq = translate(seq_translate[frame:])
    if condensed == True:
        return(protein_seq)
    else:
        formatted_seq = ''
        count = 0 - frame
        amino_acid = 0
        for base in seq:
            if base.isupper():
                count += 1
                if count % 3 == 1:
                    try:
                        formatted_seq += protein_seq[amino_acid]
                        amino_acid += 1
                    except IndexError:
                        count -= 1
                else:
                    formatted_seq += ' '
            else:
                formatted_seq += ' '
        return formatted_seq


def strip_DNA_sequence(sequence, length):
    """ Removes all characters that are not AGCTacgt from string """
    output = ''
    count = 0
    for base in sequence:
        if base in 'ACGTacgt':
            output += base
            count += 1
            if count == length:
                output += '\n'
                count = 0

    return output

def change_DNA_case(sequence, change):
    """ Converts sequence to all uppercase or all lowercase"""
    output = ''
    for base in sequence:
        if change == 'u':
            output += base.upper()
        elif change == 'l':
            output += base.lower()

    return output